/* SPDX-License-Identifier: MIT
Copyright (c) 2025 Matthias Bilger <matthias@bilger.info>
*/
body {
  font-family: sans-serif;
  line-height: 1.6;
  margin: 0;
  padding: 0;
  background-color: #f4f4f4;
  color: #333;
}

header {
  background: #666;
  color: #fff;
  padding: 1rem 0;
  text-align: center;
}

main {
  padding: 2rem;
  max-width: 800px;
  margin: 0 auto;
}

h1 {
  margin-bottom: 1rem;
}

h2 {
  margin-top: 2rem;
}

ul {
  list-style: disc;
  padding-left: 2rem;
}

a {
  color: #333;
  text-decoration: none;
}

a:hover {
  text-decoration: underline;
}

.faq-question {
  font-weight: bold;
  margin-bottom: 0.5rem;
}

.faq-answer {
  margin-bottom: 1.5rem;
}

footer {
  background: #333;
  color: #fff;
  text-align: center;
  padding: 1rem 0;
  width: 100%;
  height: 70px;
}

#readme-content,
#license-content {
  max-width: 800px;
  margin: 2em auto;
  padding: 1em;
  background-color: #fff;
  border-radius: 8px;
  box-shadow: 0 4px 8px rgba(0, 0, 0, 0.1);
  word-wrap: break-word;
}

#readme-content p,
#license-content p {
  line-height: 1.6;
  margin-bottom: 1em;
  word-wrap: break-word;
}

#readme-content code,
#license-content code {
  background-color: #e1e1e1;
  padding: 0.2em 0.4em;
  border-radius: 4px;
  word-wrap: break-word;
}

#readme-content pre,
#license-content pre {
  background-color: #e1e1e1;
  padding: 1em;
  border-radius: 8px;
  overflow-x: auto;
  word-wrap: break-word;
}

footer a {
    color: #fff;
}

footer a:hover {
    color: #ddd;
}

header::before {
  content: url('/images/mutenix_logo_finallogo_active.png');
  display: block;
  margin-bottom: 10px;
}
table tbody tr:nth-child(odd) {
    background-color: #f2f2f2;
}
table tbody tr:nth-child(even) {
    background-color: #ffffff;
}
