# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger matthias@bilger.info
# This file is intentionally left blank to mark the directory as a package.
from __future__ import annotations
