<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>Popup</title>
</head>
<body>
    <div id="mainDiv"
        style="display: flex; justify-content: center; align-items: center">
        {% include 'keypads.html' %}
    </div>
    <script>
        document.addEventListener('DOMContentLoaded', function () {
            const buttonSelection = localStorage.getItem('buttonSelection');
            activateButton(buttonSelection);
        });
    </script>
    <script src="/static/js/script.js"></script>
</body>
</html>
