<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<!-- SPDX-License-Identifier: MIT -->
<!-- Copyright (c) 2025 Matthias Bilger matthias@bilger.info -->
<html lang="en">
<head profile="http://www.w3.org/2005/10/profile">
    <link rel="icon" type="image/png" sizes="64x64" href="/favicon/64">
    <link rel="shortcut icon" type="image/png" href="/favicon/svg">
    <link rel="icon" type="image/png" sizes="32x32" href="/favicon/32">
    <link rel="icon" type="image/png" sizes="16x16" href="/favicon/16">
    <link rel="apple-touch-icon" sizes="180x180" href="/favicon/apple_touch">
    <link rel="icon" type="image/x-icon" href="/favicon.ico">
    <link rel="manifest" href="/site.webmanifest">
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <link rel="stylesheet" href="/static/css/styles.css">
    <title>{% block title %}Mutenix{% endblock %}</title>
</head>
<body>
    {% block content %}{% endblock %}
    <footer>
        <a href="/">Virtual Macropad</a> &middot; <a href="/about">About</a> &middot; <a href="/config">Config</a> &middot; <a href="/device">Device</a> &middot; <a href="/help">Help</a><br />
        <small>&copy; 2025 Matthias Bilger</small> &middot; <small>Version {{ mutenix_version }}</small>
    </footer>
</body>
</html>
